#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <string.h>

#include <getopt.h>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glut.h>

#include <vector>
#include <map>

#include <pthread.h>

using namespace std;

int lifetime_sec = 30;
int ani_tick = 1000;
int ndivisions = 0;
int pixelres = 30000;

pthread_t *sthread;
pthread_mutex_t slock;

// Pixel resolution point used for the map
typedef struct point_res {
	int x;
	int y;
	int z;
};

// Render point
typedef struct points {
	int cacheid;
    uint32_t src, dst;
    uint16_t dport;

	float x;
	float y;
	float z;
	time_t ts;

	// Color and 3d position
	float cx, cy, cz;
	float px, py, pz;
};

typedef struct net_combo {
	struct in_addr addr;
	struct in_addr mask;
	struct in_addr bcst;

	uint32_t endian_addr;
	uint32_t endian_mask;
	uint32_t endian_bcst;

	uint32_t max;

	char mintxt[128];
	char maxtxt[128];
};

vector<points> glob_points;
map<int, int> glob_rescache;

// Mouse stuff
float glob_angle = 0.0f;
float glob_anitheta = 0.0f;
int glob_movemode = 0;
int glob_motx = 0, glob_moty = 0;

// Networks we care about
net_combo glob_srcnets;
net_combo glob_dstnets;

// Point size
float glob_ptsz = 3.0;

uint32_t Adler32Checksum(const char *buf1, int len) {
	int i;
	uint32_t s1, s2;
	char *buf = (char *)buf1;
	int CHAR_OFFSET = 0;

	s1 = s2 = 0;
	for (i = 0; i < (len-4); i+=4) {
		s2 += 4*(s1 + buf[i]) + 3*buf[i+1] + 2*buf[i+2] + buf[i+3] + 
			10*CHAR_OFFSET;
		s1 += (buf[i+0] + buf[i+1] + buf[i+2] + buf[i+3] + 4*CHAR_OFFSET); 
	}

	for (; i < len; i++) {
		s1 += (buf[i]+CHAR_OFFSET); s2 += s1;
	}

	return (s1 & 0xffff) + (s2 << 16);
}

void drawtext(char *txt) {
	char *p = txt;
	while (*p != '\0') {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *p);
		p++;
	}
}

void DoomCube() {
	glPushMatrix();
	int killnum = 0;
	time_t cur = time(0);
	float cdelta = 1.0f;

	glColor3f(0.0, 1.0, 0.0);

	glRasterPos3f(-1.0, 1.0, -1.0);
	drawtext(glob_srcnets.mintxt);

	glRasterPos3f(1.0, 1.0, -1.0);
	drawtext(glob_srcnets.maxtxt);

	glRasterPos3f(1.0, -1.0, -1.0);
	drawtext(glob_dstnets.mintxt);

	glRasterPos3f(1.0, -1.0, 1.0);
	drawtext(glob_dstnets.maxtxt);

	glRasterPos3f(-1.0, 1.0, 1.0);
	drawtext("DPORT 65535");

	glRasterPos3f(-1.0, -1.0, 1.0);
	drawtext("DPORT 0");
	
	glBegin(GL_QUADS);		// Draw The Cube Using quads
	glColor3f(0.0f,1.0f,0.0f);	// Color Blue
	glVertex3f( 1.0f, 1.0f,-1.0f);	// Top Right Of The Quad (Top)
	glVertex3f(-1.0f, 1.0f,-1.0f);	// Top Left Of The Quad (Top)
	glVertex3f(-1.0f, 1.0f, 1.0f);	// Bottom Left Of The Quad (Top)
	glVertex3f( 1.0f, 1.0f, 1.0f);	// Bottom Right Of The Quad (Top)
	glColor3f(1.0f,0.5f,0.0f);	// Color Orange
	glVertex3f( 1.0f,-1.0f, 1.0f);	// Top Right Of The Quad (Bottom)
	glVertex3f(-1.0f,-1.0f, 1.0f);	// Top Left Of The Quad (Bottom)
	glVertex3f(-1.0f,-1.0f,-1.0f);	// Bottom Left Of The Quad (Bottom)
	glVertex3f( 1.0f,-1.0f,-1.0f);	// Bottom Right Of The Quad (Bottom)
	glColor3f(1.0f,0.0f,0.0f);	// Color Red	
	glVertex3f( 1.0f, 1.0f, 1.0f);	// Top Right Of The Quad (Front)
	glVertex3f(-1.0f, 1.0f, 1.0f);	// Top Left Of The Quad (Front)
	glVertex3f(-1.0f,-1.0f, 1.0f);	// Bottom Left Of The Quad (Front)
	glVertex3f( 1.0f,-1.0f, 1.0f);	// Bottom Right Of The Quad (Front)
	glColor3f(1.0f,1.0f,0.0f);	// Color Yellow
	glVertex3f( 1.0f,-1.0f,-1.0f);	// Top Right Of The Quad (Back)
	glVertex3f(-1.0f,-1.0f,-1.0f);	// Top Left Of The Quad (Back)
	glVertex3f(-1.0f, 1.0f,-1.0f);	// Bottom Left Of The Quad (Back)
	glVertex3f( 1.0f, 1.0f,-1.0f);	// Bottom Right Of The Quad (Back)
	glColor3f(0.0f,0.0f,1.0f);	// Color Blue
	glVertex3f(-1.0f, 1.0f, 1.0f);	// Top Right Of The Quad (Left)
	glVertex3f(-1.0f, 1.0f,-1.0f);	// Top Left Of The Quad (Left)
	glVertex3f(-1.0f,-1.0f,-1.0f);	// Bottom Left Of The Quad (Left)
	glVertex3f(-1.0f,-1.0f, 1.0f);	// Bottom Right Of The Quad (Left)
	glColor3f(1.0f,0.0f,1.0f);	// Color Violet
	glVertex3f( 1.0f, 1.0f,-1.0f);	// Top Right Of The Quad (Right)
	glVertex3f( 1.0f, 1.0f, 1.0f);	// Top Left Of The Quad (Right)
	glVertex3f( 1.0f,-1.0f, 1.0f);	// Bottom Left Of The Quad (Right)
	glVertex3f( 1.0f,-1.0f,-1.0f);	// Bottom Right Of The Quad (Right)
	glEnd();			// End Drawing The Cube

	glBegin(GL_LINES);              // Draw The grid using lines

	//Set increment value for number of divisions
	float step = 2 / (float)ndivisions;

	for(float i=-1+step; i<1; i+=step) {
		//Draw top grid grid horiz
		glColor3f(0.3f,0.3f,0.3f);      // Color light gray
		glVertex3f(-1.0f, 1.0f, i);
		glVertex3f( 1.0f, 1.0f, i);

		glColor3f(0.3f,0.3f,0.3f);      // Color light gray
		//Draw top grid, vertical
		glVertex3f(i, 1.0f, -1.0f);
		glVertex3f(i, 1.0f,  1.0f);

		// printf("backangle: %f\n", backangle);

		cdelta = 1.0f;

		// visible 294-78, fade-visible 284-98
		if (glob_angle >= 284 || glob_angle <= 98) {
			if (glob_angle <= 294) 
				cdelta = glob_angle / 294;
			else if (glob_angle >= 78)
				cdelta = 78 / glob_angle;

			printf("angle %d delta %f\n", glob_angle, cdelta);
			glColor3f(0.3f * cdelta, 0.3f * cdelta, 0.3f * cdelta);

			// Yellow
			glVertex3f(-1.0f, i, -1.0f);
			glVertex3f( 1.0f, i, -1.0f);

			glVertex3f(i, -1.0, -1.0f);
			glVertex3f(i, 1.0,  -1.0f);
		}

		if (glob_angle >= 111 && glob_angle <= 251) {
			glColor3f(0.3f * cdelta, 0.3f * cdelta, 0.3f * cdelta);

			// red
			glVertex3f(-1.0f, i, 1.0f);
			glVertex3f( 1.0f, i, 1.0f);

			glVertex3f(i, -1.0, 1.0f);
			glVertex3f(i, 1.0,  1.0f);
		}

		if (glob_angle >= 197 && glob_angle <= 336) {
			glColor3f(0.3f * cdelta, 0.3f * cdelta, 0.3f * cdelta);

			// Blue
			glVertex3f(-1.0f, -1.0f, i);
			glVertex3f(-1.0f,  1.0f, i);

			glVertex3f(-1.0f, i, -1.0f);
			glVertex3f(-1.0f, i,  1.0f);
		}

		if (glob_angle >= 19 && glob_angle <= 163) {
			glColor3f(0.3f * cdelta, 0.3f * cdelta, 0.3f * cdelta);

			// Purple
			glVertex3f(1.0f, -1.0f, i);
			glVertex3f(1.0f,  1.0f, i);

			glVertex3f(1.0f, i, -1.0f);
			glVertex3f(1.0f, i,  1.0f);
		}

		//Draw bottom grid horiz
		glVertex3f(-1.0f, -1.0f, i);
		glVertex3f( 1.0f, -1.0f, i);

		//Draw bottom grid, vertical
		glVertex3f(i, -1.0f, -1.0f);
		glVertex3f(i, -1.0f,  1.0f);
	}
	glEnd();			//End drawing grids

    // Set the point size
    glPointSize(glob_ptsz);

	glBegin(GL_POINTS);

	for (unsigned int x = 0; x < glob_points.size(); x++) {
        int dtime;

		if ((dtime = cur - glob_points[x].ts) > lifetime_sec) {
			killnum = x;
			continue;
		}

		
#if 1
        float cdelta = 0.0f;
        if (dtime > lifetime_sec / 2) {
            cdelta = ((float) 0.4f * ((float) dtime / (float) lifetime_sec));
        }
#endif

#if 0
        float cdelta = 1.0f;
		cdelta -= (float) ((float) dtime / (float) lifetime_sec);

		printf("%d %d - %f\n", dtime, lifetime_sec, cdelta);
#endif
       
#if 1
		glColor3f(glob_points[x].cx - cdelta, 
				  glob_points[x].cy - cdelta, 
				  glob_points[x].cz - cdelta);
#endif

#if 0
		glEnable(GL_BLEND); 
		glColor4f(glob_points[x].cx, 
				  glob_points[x].cy, 
				  glob_points[x].cz,
				  cdelta);
		glBlendFunc(GL_ONE, GL_ONE);
#endif

		glVertex3f(glob_points[x].px, 
				   glob_points[x].py, 
				   glob_points[x].pz);
	}
	glEnd();
	
	glPopMatrix();

	if (killnum != 0) {
		for (unsigned int z = 0; z < killnum; z++) 
			glob_rescache.erase(glob_rescache.find(glob_points[z].cacheid));

		glob_points.erase(glob_points.begin(), glob_points.begin() + killnum);
	}
}

int ParseIpset(char *txt, char *clstype, net_combo *ret_combo) {
	short int ip0, ip1, ip2, ip3;
	short int nm0, nm1, nm2, nm3;

	if (sscanf(txt, "%hu.%hu.%hu.%hu/%hu.%hu.%hu.%hu",
			   &ip0, &ip1, &ip2, &ip3,
			   &nm0, &nm1, &nm2, &nm3) != 8) {
		fprintf(stderr, "Error:  Invalid network, should be of the "
				"form 1.2.3.4/255.255.255.0\n");
		return -1;
	}

	ret_combo->addr.s_addr = 0;
	ret_combo->addr.s_addr |= (ip3 << 24);
	ret_combo->addr.s_addr |= (ip2 << 16);
	ret_combo->addr.s_addr |= (ip1 << 8);
	ret_combo->addr.s_addr |= (ip0 << 0);

	ret_combo->mask.s_addr = 0;
	ret_combo->mask.s_addr |= (nm3 << 24);
	ret_combo->mask.s_addr |= (nm2 << 16);
	ret_combo->mask.s_addr |= (nm1 << 8);
	ret_combo->mask.s_addr |= (nm0 << 0);

	ret_combo->bcst.s_addr = 
		(ret_combo->addr.s_addr | ~(ret_combo->mask.s_addr));

	// Handle initializing to "all addresses"
	if (ret_combo->mask.s_addr == 0) {
		ret_combo->mask.s_addr = 0x0;
		ret_combo->bcst.s_addr = 0xFFFFFFFF;
		ret_combo->endian_addr = 0;
		ret_combo->endian_bcst = 0xFFFFFFFF;
		ret_combo->endian_mask = 0x0;
		ret_combo->max = 0xFFFFFFFF;
	} else {
		ret_combo->endian_addr = ntohl(ret_combo->addr.s_addr);
		ret_combo->endian_mask = ntohl(ret_combo->mask.s_addr);
		ret_combo->endian_bcst = ntohl(ret_combo->bcst.s_addr);

		ret_combo->max = (ret_combo->endian_bcst & ~ret_combo->endian_mask);
	}

	snprintf(ret_combo->mintxt, 128, "%s %s", clstype, inet_ntoa(ret_combo->addr));
	snprintf(ret_combo->maxtxt, 128, "%s %s", clstype, inet_ntoa(ret_combo->bcst));

	return 1;
}

void PointMagic(uint32_t src, uint32_t dst, uint16_t dport, int timesec) {
	points pt;
	point_res ptr;
	int cs;

    // Check the last 10 points and see if we just added one, don't duplicate
    // if we did.
    int pos;
    if ((pos = glob_points.size() - 10) < 0)
        pos = 0;

    for (pos; (unsigned int) pos < glob_points.size(); pos++) {
        if (glob_points[pos].src == src &&
            glob_points[pos].dst == dst &&
            glob_points[pos].dport == dport)
            return;
    }
    
	pt.ts = timesec;

    pt.src = src;
    pt.dst = dst;
    pt.dport = dport;

	uint32_t smax = glob_srcnets.max;
	uint32_t dmax = glob_dstnets.max;

	// Does it fall within our network ranges?  Discard it if it doesn't.
	if (((src & glob_srcnets.mask.s_addr) != (glob_srcnets.addr.s_addr) &&
		 glob_srcnets.addr.s_addr != 0) ||
		((dst & glob_dstnets.mask.s_addr) != (glob_dstnets.addr.s_addr) &&
		 glob_dstnets.addr.s_addr != 0)) {

		// printf("%u & %u = %u not %u\n", src, glob_srcnets.mask.s_addr, src & glob_srcnets.mask.s_addr, glob_srcnets.addr.s_addr);

		return;
	}

	src = (ntohl(src) & ~(glob_srcnets.endian_mask));
	dst = (ntohl(dst) & ~(glob_dstnets.endian_mask));

	// Figure out the resolution cache values and check
	ptr.x = (int) ((float) ((float) src / (float) smax) * pixelres);
	ptr.y = (int) ((float) ((float) dport / 65535.0f) * pixelres);
	ptr.z = (int) ((float) ((float) dst / (float) dmax) * pixelres);

	// Checksum it to a single int
	cs = Adler32Checksum((const char *) &ptr, sizeof(point_res));

	// Check
	if (glob_rescache.find(cs) != glob_rescache.end()) {
		// printf("debug - throwing out cached point %d,%d,%d %d\n", ptr.x, ptr.y, ptr.z, cs);
		return;
	}

	// Cache it
	glob_rescache[cs] = 1;

	pt.cacheid = cs;

	pt.x = (float)((float) src / (float) smax);
	pt.y = (float)((float) dport / 65535.0f);
	pt.z = (float)((float) dst / (float) dmax);

	pt.cx = (0.9 * pt.x) + 0.1;
	pt.cy = (0.9 * pt.y) + 0.1;
	pt.cz = (0.9 * pt.z) + 0.1;

	pt.px = (2.0f * pt.x) - 1.0f;
	pt.py = (2.0f * pt.y) - 1.0f;
	pt.pz = (2.0f * pt.z) - 1.0f;

	glob_points.push_back(pt);
}

void ParseStdinIpgroup(char *txt) {
	short int ip0, ip1, ip2, ip3;
	short int np0, np1, np2, np3;
	int dport;

	uint32_t src;
	uint32_t dst;

	if (sscanf(txt, "%hu.%hu.%hu.%hu %hu.%hu.%hu.%hu %d",
			   &ip0, &ip1, &ip2, &ip3,
			   &np0, &np1, &np2, &np3,
			   &dport) != 9) {
		fprintf(stderr, "Invalid stdin line, should be 'srcip dstip port'\n");
		return;
	}

	src = 0;
	src |= (ip3 << 24);
	src |= (ip2 << 16);
	src |= (ip1 << 8);
	src |= (ip0 << 0);

	dst = 0;
	dst |= (np3 << 24);
	dst |= (np2 << 16);
	dst |= (np1 << 8);
	dst |= (np0 << 0);

	PointMagic(src, dst, dport, time(0));
}


void mouse(int b, int m, int x, int y) {
    // Left button down sets move mode
    if (b == 1) {
        glob_anitheta = 0;
        if (m) {
            glob_movemode = 1;
            glob_motx = x;
            glob_moty = y;
        } else {
            glob_movemode = 0;
        }
    } else if (b == 3) {
        if (m) {
            glob_movemode = 2;
            glob_motx = x;
            glob_moty = y;
        } else {
            glob_movemode = 0;
        }
    }
}

void motion(int x, int y) {
    if (glob_movemode) {
        int dx = (x - glob_motx);
        glob_motx = x;

        glob_angle += dx;

		if (glob_angle < 0)
			glob_angle += 360;
		else if (glob_angle > 360)
			glob_angle -= 360;

        if (glob_movemode == 2)
            glob_anitheta = dx;
    }
}

void Usage(char *argv) {
	printf("Usage: %s [OPTION]\n", argv);
	printf(" -f, --fullscreen          Fullscreen video\n"
		   " -w, --window-size <x,y>   Window/screen size\n"
		   " -s, --srcnets   <net/msk> Network range for src\n"
		   " -d, --dstnets   <net/msk> Network range for dst\n"
           " -l, --lifetime    <x>     Lifetime in seconds of a point\n"
           " -a, --anitick     <x>     Tick in animation - less is smoother but more CPU\n"
           " -P, --point-size  <x[.y]> Size of points in the cube\n"
		   " -n, --ndivisions  <x>     Number of graph lines\n"
		   " -r, --resolution  <x>     Number of samples for each cube side\n"
		   "   --initial-theta <x[.y]> Initial rotation theta\n"
		  );
}

void init_gl_screen(int screenx, int screeny) {
	glViewport(0, 0, screenx, screeny);
	glClearDepth(1.0);
	glDepthFunc(GL_LEQUAL);

	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(40, screenx / screeny, 1.5, 20.0);
	glMatrixMode(GL_MODELVIEW);

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
}

int main(int argc, char *argv[]) {
	SDL_Event event;
	SDL_Surface *screen;
	float depth = -3.0;
	int running = 1;
	char errstr[1024];

	// cmdline stuff
	int fullscreen = 0;
	int screenx = 640, screeny = 480;
	int option_idx = 0;

	static struct option main_longopt[] = {
		{ "fullscreen", no_argument, 0, 'f' },
		{ "window-size", required_argument, 0, 'w' },
		{ "srcnets", required_argument, 0, 's' },
		{ "dstnets", required_argument, 0, 'd' },
        { "lifetime", required_argument, 0, 'l' },
        { "anitick", required_argument, 0, 'a' },
        { "point-size", required_argument, 0, 'P' },
	    { "ndivisions", required_argument, 0, 'n' },
		{ "resolution", required_argument, 0, 'r' },
		{ "initial-theta", required_argument, 0, 200 },
		{ "help", no_argument, 0, 'h' },
		{ 0, 0, 0, 0 }
	};

	// Init glob addr stuff
	ParseIpset("0.0.0.0/0.0.0.0", "SRC", &glob_srcnets);
	ParseIpset("0.0.0.0/0.0.0.0", "DST", &glob_dstnets);

	while (1) {
		int r = getopt_long(argc, argv, "fw:s:d:l:a:P:hn:r:", main_longopt, &option_idx);
		if (r < 0) break;
		switch (r) {
			case 'f':
				fullscreen = 1;
				break;
			case 'w':
				if (sscanf(optarg, "%d,%d", &screenx, &screeny) != 2) {
					fprintf(stderr, "Error: Screen size must be X,Y\n");
					exit(1);
				}
				break;
			case 's':
				if (ParseIpset(optarg, "SRC", &glob_srcnets) < 0)
					exit(1);
				break;
			case 'd':
				if (ParseIpset(optarg, "DST", &glob_dstnets) < 0)
					exit(1);
				break;
            case 'l':
                if (sscanf(optarg, "%d", &lifetime_sec) != 1) {
                    fprintf(stderr, "Error:  Lifetime must be seconds\n");
                    exit(1);
                }
                break;
            case 'a':
                if (sscanf(optarg, "%d", &ani_tick) != 1) {
                    fprintf(stderr, "Error:  anitick must be a number\n");
                    exit(1);
                }
                break;
            case 'P':
                if (sscanf(optarg, "%f", &glob_ptsz) != 1) {
                    fprintf(stderr, "Error:  point size must be a float (x[.y])\n");
                    exit(1);
                }
                break;
			case 'n':
				if (sscanf(optarg, "%d", &ndivisions) != 1) {
					fprintf(stderr, "Error:  Grid divisions must be a number\n");
					exit(1);
				}
				break;
			case 'r':
				if (sscanf(optarg, "%d", &pixelres) != 1) {
					fprintf(stderr, "Error:  Resolution must be an integer\n");
					exit(1);
				}
				break;
            case 200:
                if (sscanf(optarg, "%f", &glob_anitheta) != 1) {
                    fprintf(stderr, "Error:  initial-theta must be a number\n");
                    exit(1);
                }
                break;
			case 'h':
				Usage(argv[0]);
				exit(1);
		}
	}

	if (SDL_Init(SDL_INIT_VIDEO) != 0) {
		fprintf(stderr, "Error: Unable to initialize SDL video: %s\n",
				SDL_GetError());
		SDL_Quit();
		exit(1);
	}

	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);		//Use at least 5 bits of Red
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);	//Use at least 5 bits of Green
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);		//Use at least 5 bits of Blue
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);	
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);	//Enable double buffering

	if (fullscreen)
		screen = SDL_SetVideoMode(screenx, screeny, 16, 
								  SDL_OPENGLBLIT | SDL_FULLSCREEN);
	else
		screen = SDL_SetVideoMode(screenx, screeny, 16, 
								  SDL_OPENGLBLIT);

	if (screen == NULL) {
		fprintf(stderr, "Error: Unable to set video mode: %s\n",
				SDL_GetError());
		return 1;
	}

	init_gl_screen(screenx, screeny);

	glutInit(&argc, argv);

	fd_set rset;
	int maxfd = 0;
	struct timeval tm;

	while (running) {
		FD_ZERO(&rset);
		maxfd = 0;

		FD_SET(maxfd, &rset);
		struct timeval leading, tailing;

		gettimeofday(&leading, NULL);

		tm.tv_sec = 0;
		tm.tv_usec = ani_tick;

		if (select(maxfd + 1, &rset, NULL, NULL, &tm) < 0) {
			if (errno != EINTR) {
				fprintf(stderr, "Main select loop failed\n");
				exit(1);
			}
		}

		if (FD_ISSET(maxfd, &rset)) {
			char stdinchr[512];
			fgets(stdinchr, 512, stdin);
			ParseStdinIpgroup(stdinchr);
		}

		// Throttle speed down in high-packet situations
		gettimeofday(&tailing, NULL);
		if (leading.tv_sec != tailing.tv_sec ||
			tailing.tv_usec - leading.tv_usec > ani_tick) {

			// Clear screen, color
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			// Change where we look at
			gluLookAt(0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

			// The cubes are transformed as they're drawn spatially
			glPushMatrix();

			glob_angle += (glob_anitheta / 10);
			if (glob_angle > 360)
				glob_angle -= 360;
			else if (glob_angle < 0)
				glob_angle += 360;

			glRotatef(glob_angle, 0.0, 1.0, 0.0);
			DoomCube();

			glPopMatrix();

			glFlush();

			SDL_GL_SwapBuffers();
		}

		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_ESCAPE) {
						running = 0;
						break;
					}

					break;

				case SDL_KEYUP:
					if (event.key.keysym.sym == 'f') {
						SDL_FreeSurface(screen);

						if (fullscreen) {
							screen = 
								SDL_SetVideoMode(screenx, screeny, 16, 
												 SDL_OPENGLBLIT);
							fullscreen = 0;
						} else {
							screen = 
								SDL_SetVideoMode(screenx, screeny, 16, 
												 SDL_OPENGLBLIT | SDL_FULLSCREEN);
							fullscreen = 1;
						}

						if (screen == NULL) {
							fprintf(stderr, "Error: Unable to set video mode: %s\n",
									SDL_GetError());
							return 1;
						}

						init_gl_screen(screenx, screeny);
					}
					
					break;
                case SDL_MOUSEBUTTONDOWN:
                case SDL_MOUSEBUTTONUP:
                    mouse(event.button.button, event.button.state, 
						  event.button.x, event.button.y);
                    break;
                case SDL_MOUSEMOTION:
                    motion(event.motion.x, event.motion.y);
                    break;
				case SDL_QUIT:
					running = 0;
					break;
			}
		}

	}

	SDL_Quit();
	exit(0);
}
